@@ -161,7 +161,7 @@ class SyncObj(object):
         self.__raftLog = createJournal(self.__conf.journalFile)
         if len(self.__raftLog) == 0:
             self.__raftLog.add(_bchr(_COMMAND_TYPE.NO_OP), 1, self.__raftCurrentTerm)
-        self.__raftCommitIndex = 1
+        self.__raftCommitIndex = self.__raftLog.getRaftCommitIndex()
         self.__raftLastApplied = 1
         self.__raftNextIndex = {}
         self.__lastResponseTime = {}
@@ -180,6 +180,7 @@ class SyncObj(object):
         self.__onTickCallbacksLock = threading.Lock()
 
         self.__startTime = monotonicTime()
+        self.__numOneSecondDumps = 0
         globalDnsResolver().setTimeouts(self.__conf.dnsCacheTime, self.__conf.dnsFailCacheTime)
         globalDnsResolver().setPreferredAddrFamily(self.__conf.preferredAddrType)
         self.__serializer = Serializer(self.__conf.fullDumpFile,
@@ -200,6 +201,10 @@ class SyncObj(object):
         self.__transport.setOnMessageReceivedCallback(self.__onMessageReceived)
         self.__transport.setOnReadonlyNodeConnectedCallback(self.__onReadonlyNodeConnected)
         self.__transport.setOnReadonlyNodeDisconnectedCallback(self.__onReadonlyNodeDisconnected)
+        self.__transport.setOnUtilityMessageCallback('status', self._getStatus)
+        self.__transport.setOnUtilityMessageCallback('add', self._addNodeToCluster)
+        self.__transport.setOnUtilityMessageCallback('remove', self._removeNodeFromCluster)
+        self.__transport.setOnUtilityMessageCallback('set_version', self._setCodeVersion)
 
         self._methodToID = {}
         self._idToMethod = {}
@@ -324,7 +329,7 @@ class SyncObj(object):
 
         :param newVersion: new code version
         :type int
-        :param callback: will be called on cussess or fail
+        :param callback: will be called on success or fail
         :type callback: function(`FAIL_REASON <#pysyncobj.FAIL_REASON>`_, None)
         """
         assert isinstance(newVersion, int)
@@ -366,11 +371,18 @@ class SyncObj(object):
             node = self.__nodeClass(node)
         self._applyCommand(pickle.dumps(['rem', node.id, node]), callback, _COMMAND_TYPE.MEMBERSHIP)
 
-    def _addNodeToCluster(self, node, callback=None):
-        self.addNodeToCluster(node, callback)
+    def _setCodeVersion(self, args, callback):
+        self.setCodeVersion(args[0], callback)
 
-    def _removeNodeFromCluster(self, node, callback=None):
-        self.removeNodeFromCluster(node, callback)
+    def _addNodeToCluster(self, args, callback):
+        self.addNodeToCluster(args[0], callback)
+
+    def _removeNodeFromCluster(self, args, callback):
+        node = args[0]
+        if node == self.__selfNode.address:
+            callback(None, FAIL_REASON.REQUEST_DENIED)
+        else:
+            self.removeNodeFromCluster(node, callback)
 
     def __onSetCodeVersion(self, newVersion):
         methods = [m for m in dir(self) if callable(getattr(self, m)) and\
@@ -529,6 +541,7 @@ class SyncObj(object):
 
         if not self.__transport.ready:
             time.sleep(timeToWait)
+            self.__applyLogEntries()
             return
 
         if self.__needLoadDumpFile:
@@ -536,6 +549,11 @@ class SyncObj(object):
                 self.__loadDumpFile(clearJournal=False)
             self.__needLoadDumpFile = False
 
+        workTime = monotonicTime() - self.__startTime
+        if workTime > self.__numOneSecondDumps:
+            self.__numOneSecondDumps += 1
+            self.__raftLog.onOneSecondTimer()
+
         if self.__raftState in (_RAFT_STATE.FOLLOWER, _RAFT_STATE.CANDIDATE) and self.__selfNode is not None:
             if self.__raftElectionDeadline < monotonicTime() and self.__connectedToAnyone():
                 self.__raftElectionDeadline = monotonicTime() + self.__generateRaftTimeout()
@@ -564,6 +582,7 @@ class SyncObj(object):
                         count += 1
                 if count > (len(self.__otherNodes) + 1) / 2:
                     self.__raftCommitIndex = nextCommitIndex
+                    self.__raftLog.setRaftCommitIndex(self.__raftCommitIndex)
                 else:
                     break
             self.__leaderCommitIndex = self.__raftCommitIndex
@@ -576,30 +595,7 @@ class SyncObj(object):
                 self.__setState(_RAFT_STATE.FOLLOWER)
                 self.__raftLeader = None
 
-        needSendAppendEntries = False
-
-        if self.__raftCommitIndex > self.__raftLastApplied:
-            count = self.__raftCommitIndex - self.__raftLastApplied
-            entries = self.__getEntries(self.__raftLastApplied + 1, count)
-            for entry in entries:
-                try:
-                    currentTermID = entry[2]
-                    subscribers = self.__commandsWaitingCommit.pop(entry[1], [])
-                    res = self.__doApplyCommand(entry[0])
-                    for subscribeTermID, callback in subscribers:
-                        if subscribeTermID == currentTermID:
-                            callback(res, FAIL_REASON.SUCCESS)
-                        else:
-                            callback(None, FAIL_REASON.DISCARDED)
-
-                    self.__raftLastApplied += 1
-                except SyncObjExceptionWrongVer as e:
-                    logging.error('request to switch to unsupported code version (self version: %d, requested version: %d)' %
-                        (self.__selfCodeVersion, e.ver))
-
-
-            if not self.__conf.appendEntriesUseBatch:
-                needSendAppendEntries = True
+        needSendAppendEntries = self.__applyLogEntries()
 
         if self.__raftState == _RAFT_STATE.LEADER:
             if monotonicTime() > self.__newAppendEntriesTime or needSendAppendEntries:
@@ -619,6 +615,34 @@ class SyncObj(object):
 
         self._poller.poll(timeToWait)
 
+    def __applyLogEntries(self):
+        needSendAppendEntries = False
+
+        if self.__raftCommitIndex > self.__raftLastApplied:
+            count = self.__raftCommitIndex - self.__raftLastApplied
+            entries = self.__getEntries(self.__raftLastApplied + 1, count)
+            for entry in entries:
+                try:
+                    currentTermID = entry[2]
+                    subscribers = self.__commandsWaitingCommit.pop(entry[1], [])
+                    res = self.__doApplyCommand(entry[0])
+                    for subscribeTermID, callback in subscribers:
+                        if subscribeTermID == currentTermID:
+                            callback(res, FAIL_REASON.SUCCESS)
+                        else:
+                            callback(None, FAIL_REASON.DISCARDED)
+
+                    self.__raftLastApplied += 1
+                except SyncObjExceptionWrongVer as e:
+                    logging.error(
+                        'request to switch to unsupported code version (self version: %d, requested version: %d)' %
+                        (self.__selfCodeVersion, e.ver))
+
+            if not self.__conf.appendEntriesUseBatch:
+                needSendAppendEntries = True
+
+        return needSendAppendEntries
+
     def addOnTickCallback(self, callback):
         with self.__onTickCallbacksLock:
             self.__onTickCallbacks.append(callback)
@@ -631,25 +655,76 @@ class SyncObj(object):
                 # callback not in list, ignore
                 pass
 
+    def isNodeConnected(self, node):
+        """
+        Checks if the given node is connected
+        :param node: node to check
+        :type node: Node
+        :rtype: bool
+        """
+        return node in self.__connectedNodes
+
+    @property
+    def selfNode(self):
+        """
+        :rtype: Node
+        """
+        return self.__selfNode
+
+    @property
+    def otherNodes(self):
+        """
+        :rtype: set of Node
+        """
+        return self.__otherNodes.copy()
+
+    @property
+    def readonlyNodes(self):
+        """
+        :rtype: set of Node
+        """
+        return self.__readonlyNodes.copy()
+
+    @property
+    def raftLastApplied(self):
+        """
+        :rtype: int
+        """
+        return self.__raftLastApplied
+
+    @property
+    def raftCommitIndex(self):
+        """
+        :rtype: int
+        """
+        return self.__raftCommitIndex
+
+    @property
+    def raftCurrentTerm(self):
+        """
+        :rtype: int
+        """
+        return self.__raftCurrentTerm
+
     def getStatus(self):
         """Dumps different debug info about cluster to dict and return it"""
 
         status = {}
         status['version'] = VERSION
         status['revision'] = 'deprecated'
-        status['self'] = self.__selfNode
+        status['self'] = self.selfNode
         status['state'] = self.__raftState
         status['leader'] = self.__raftLeader
         status['partner_nodes_count'] = len(self.__otherNodes)
         for node in self.__otherNodes:
-            status['partner_node_status_server_' + node.id] = 2 if node in self.__connectedNodes else 0
+            status['partner_node_status_server_' + node.id] = 2 if self.isNodeConnected(node) else 0
         status['readonly_nodes_count'] = len(self.__readonlyNodes)
         for node in self.__readonlyNodes:
-            status['readonly_node_status_server_' + node.id] = 2 if node in self.__connectedNodes else 0
+            status['readonly_node_status_server_' + node.id] = 2 if self.isNodeConnected(node) else 0
         status['log_len'] = len(self.__raftLog)
-        status['last_applied'] = self.__raftLastApplied
-        status['commit_idx'] = self.__raftCommitIndex
-        status['raft_term'] = self.__raftCurrentTerm
+        status['last_applied'] = self.raftLastApplied
+        status['commit_idx'] = self.raftCommitIndex
+        status['raft_term'] = self.raftCurrentTerm
         status['next_node_idx_count'] = len(self.__raftNextIndex)
         for node, idx in iteritems(self.__raftNextIndex):
             status['next_node_idx_server_' + node.id] = idx
@@ -662,8 +737,8 @@ class SyncObj(object):
         status['enabled_code_version'] = self.__enabledCodeVersion
         return status
 
-    def _getStatus(self):
-        return self.getStatus()
+    def _getStatus(self, args, callback):
+        callback(self.getStatus(), None)
 
     def printStatus(self):
         """Dumps different debug info about cluster to default logger"""
@@ -695,6 +770,14 @@ class SyncObj(object):
             if callback is not None:
                 callback(oldVer, ver)
             return
+
+        #  This is required only after node restarts and apply journal
+        # for normal case it is already done earlier and calls will be ignored
+        clusterChangeRequest = self.__parseChangeClusterRequest(command)
+        if clusterChangeRequest is not None:
+             self.__doChangeCluster(clusterChangeRequest)
+             return
+
         if commandType != _COMMAND_TYPE.REGULAR:
             return
         command = pickle.loads(command[1:])
@@ -815,6 +898,7 @@ class SyncObj(object):
                     self.__sendNextNodeIdx(node, success=True)
 
             self.__raftCommitIndex = min(leaderCommitIndex, self.__getCurrentLogIndex())
+            self.__raftLog.setRaftCommitIndex(self.__raftCommitIndex)
 
         if message['type'] == 'apply_command':
             if 'request_id' in message:
@@ -1159,7 +1243,7 @@ class SyncObj(object):
         if adding:
             newNode = requestNode
             # Node already exists in cluster
-            if newNode is self.__selfNode or newNode in self.__otherNodes:
+            if newNode == self.__selfNode or newNode in self.__otherNodes:
                 return False
             self.__otherNodes.add(newNode)
             self.__raftNextIndex[newNode] = self.__getCurrentLogIndex() + 1
@@ -1170,7 +1254,7 @@ class SyncObj(object):
             return True
         else:
             oldNode = requestNode
-            if oldNode is self.__selfNode:
+            if oldNode == self.__selfNode:
                 return False
             if oldNode not in self.__otherNodes:
                 return False
