@@ -7,6 +7,7 @@ import threading
 import sys
 import pysyncobj.pickle as pickle
 import pysyncobj.dns_resolver as dns_resolver
+import platform
 
 if sys.version_info >= (3, 0):
     xrange = range
@@ -15,7 +16,8 @@ import functools
 import struct
 import logging
 from pysyncobj import SyncObj, SyncObjConf, replicated, FAIL_REASON, _COMMAND_TYPE, \
-    createJournal, HAS_CRYPTO, replicated_sync, Utility, SyncObjException, SyncObjConsumer, _RAFT_STATE
+    createJournal, HAS_CRYPTO, replicated_sync, SyncObjException, SyncObjConsumer, _RAFT_STATE
+from pysyncobj.syncobj_admin import executeAdminCommand
 from pysyncobj.batteries import ReplCounter, ReplList, ReplDict, ReplSet, ReplLockManager, ReplQueue, ReplPriorityQueue
 from pysyncobj.node import TCPNode
 from collections import defaultdict
@@ -159,14 +161,8 @@ def singleTickFunc(o, timeToTick, interval, stopFunc):
                 break
 
 
-def utilityTickFunc(args, currRes, key, timeToTick):
-    u = Utility(args)
-    currTime = time.time()
-    finishTime = currTime + timeToTick
-    while time.time() < finishTime:
-        if u.getResult() is not None:
-            currRes[key] = u.getResult()
-            break
+def utilityTickFunc(args, currRes, key):
+    currRes[key] = executeAdminCommand(args)
 
 
 def doSyncObjAdminTicks(objects, arguments, timeToTick, currRes, interval=0.05, stopFunc=None):
@@ -177,7 +173,7 @@ def doSyncObjAdminTicks(objects, arguments, timeToTick, currRes, interval=0.05,
         t1.start()
         objThreads.append(t1)
         if arguments.get(o) is not None:
-            t2 = threading.Thread(target=utilityTickFunc, args=(arguments[o], currRes, o, timeToTick))
+            t2 = threading.Thread(target=utilityTickFunc, args=(arguments[o], currRes, o))
             t2.start()
             utilityThreads.append(t2)
     for t in objThreads:
@@ -615,7 +611,7 @@ def test_checkBigStorage():
 
     removeFiles(dumpFiles)
 
-
+@pytest.mark.skipif(sys.platform == "win32" or platform.python_implementation() != 'CPython', reason="does not run on windows or pypy")
 def test_encryptionCorrectPassword():
     assert HAS_CRYPTO
 
@@ -639,10 +635,22 @@ def test_encryptionCorrectPassword():
     assert o1.getCounter() == 350
     assert o2.getCounter() == 350
 
+    for conn in list(o1._SyncObj__transport._connections.values()) + list(o2._SyncObj__transport._connections.values()):
+        conn.disconnect()
+
+    doTicks(objs, 10)
+
+    o1.addValue(100)
+
+    doTicks(objs, 10, stopFunc=lambda: o1.getCounter() == 450 and o2.getCounter() == 450)
+
+    assert o1.getCounter() == 450
+    assert o2.getCounter() == 450
+
     o1._destroy()
     o2._destroy()
 
-
+@pytest.mark.skipif(platform.python_implementation() != 'CPython', reason="does not have crypto on pypy")
 def test_encryptionWrongPassword():
     assert HAS_CRYPTO
 
@@ -709,6 +717,7 @@ def _checkSameLeader2(objs):
 def test_randomTest1():
     journalFiles = [getNextJournalFile(), getNextJournalFile(), getNextJournalFile()]
     removeFiles(journalFiles)
+    removeFiles([e + '.meta' for e in journalFiles])
 
     random.seed(12)
 
@@ -779,6 +788,7 @@ def test_randomTest1():
         raise AssertionError('Values not equal')
 
     removeFiles(journalFiles)
+    removeFiles([e + '.meta' for e in journalFiles])
 
 
 # Ensure that raftLog after serialization is the same as in serialized data
@@ -949,7 +959,7 @@ def test_doChangeClusterUT2():
 
     assert o1._isReady() == o2._isReady() == o3._isReady() == True
     o3.addValue(50)
-    o2._addNodeToCluster(a[3])
+    o2.addNodeToCluster(a[3])
 
     success = False
     for i in xrange(10):
@@ -961,7 +971,7 @@ def test_doChangeClusterUT2():
         if res:
             success = True
             break
-        o2._addNodeToCluster(a[3])
+        o2.addNodeToCluster(a[3])
 
     assert success
 
@@ -982,6 +992,7 @@ def test_journalTest1():
     journalFiles = [getNextJournalFile(), getNextJournalFile()]
     removeFiles(dumpFiles)
     removeFiles(journalFiles)
+    removeFiles([e + '.meta' for e in journalFiles])
 
     random.seed(42)
 
@@ -1065,11 +1076,13 @@ def test_journalTest1():
 
     removeFiles(dumpFiles)
     removeFiles(journalFiles)
+    removeFiles([e + '.meta' for e in journalFiles])
 
 
 def test_journalTest2():
     journalFiles = [getNextJournalFile()]
     removeFiles(journalFiles)
+    removeFiles([e + '.meta' for e in journalFiles])
 
     removeFiles(journalFiles)
     journal = createJournal(journalFiles[0])
@@ -1097,6 +1110,51 @@ def test_journalTest2():
     assert journal[0] == (b'cmd2', 2, 0)
     journal._destroy()
     removeFiles(journalFiles)
+    removeFiles([e + '.meta' for e in journalFiles])
+
+
+def test_applyJournalAfterRestart():
+    dumpFiles = [getNextDumpFile(), getNextDumpFile()]
+    journalFiles = [getNextJournalFile(), getNextJournalFile()]
+    removeFiles(dumpFiles)
+    removeFiles(journalFiles)
+    removeFiles([e + '.meta' for e in journalFiles])
+
+    random.seed(42)
+
+    a = [getNextAddr(), getNextAddr()]
+
+    o1 = TestObj(a[0], [a[1]], TEST_TYPE.JOURNAL_1, dumpFile=dumpFiles[0], journalFile=journalFiles[0])
+    o2 = TestObj(a[1], [a[0]], TEST_TYPE.JOURNAL_1, dumpFile=dumpFiles[1], journalFile=journalFiles[1])
+    objs = [o1, o2]
+    doTicks(objs, 10, stopFunc=lambda: o1._isReady() and o2._isReady())
+
+    assert o1._getLeader().address in a
+    assert o1._getLeader() == o2._getLeader()
+
+    o1.addValue(150)
+    o2.addValue(200)
+
+    doTicks(objs, 10, stopFunc=lambda: o1.getCounter() == 350 and o2.getCounter() == 350)
+
+    assert o1.getCounter() == 350
+    assert o2.getCounter() == 350
+    doTicks(objs, 2)
+
+    o1._destroy()
+    o2._destroy()
+
+    removeFiles(dumpFiles)
+
+    o1 = TestObj(a[0], [a[1]], TEST_TYPE.JOURNAL_1, dumpFile=dumpFiles[0], journalFile=journalFiles[0])
+    objs = [o1]
+    doTicks(objs, 10, o1.getCounter() == 350)
+
+    assert o1.getCounter() == 350
+
+    removeFiles(dumpFiles)
+    removeFiles(journalFiles)
+    removeFiles([e + '.meta' for e in journalFiles])
 
 
 def test_autoTick1():
@@ -1205,19 +1263,19 @@ def test_largeCommands():
 
     removeFiles(dumpFiles)
 
-
+@pytest.mark.skipif(platform.python_implementation() != 'CPython', reason="does not have crypto on pypy")
 def test_readOnlyNodes():
     random.seed(12)
 
     a = [getNextAddr(), getNextAddr(), getNextAddr()]
 
-    o1 = TestObj(a[0], [a[1], a[2]])
-    o2 = TestObj(a[1], [a[2], a[0]])
-    o3 = TestObj(a[2], [a[0], a[1]])
+    o1 = TestObj(a[0], [a[1], a[2]], password='123')
+    o2 = TestObj(a[1], [a[2], a[0]], password='123')
+    o3 = TestObj(a[2], [a[0], a[1]], password='123')
     objs = [o1, o2, o3]
 
-    b1 = TestObj(None, [a[0], a[1], a[2]])
-    b2 = TestObj(None, [a[0], a[1], a[2]])
+    b1 = TestObj(None, [a[0], a[1], a[2]], password='123')
+    b2 = TestObj(None, [a[0], a[1], a[2]], password='123')
 
     roObjs = [b1, b2]
 
@@ -1308,6 +1366,7 @@ def test_readOnlyNodes():
     b2._destroy()
 
 
+@pytest.mark.skipif(platform.python_implementation() != 'CPython', reason="does not have crypto on pypy")
 def test_syncobjAdminStatus():
     assert HAS_CRYPTO
 
@@ -1326,8 +1385,8 @@ def test_syncobjAdminStatus():
     assert o1._isReady()
     assert o2._isReady()
 
-    status1 = o1._getStatus()
-    status2 = o2._getStatus()
+    status1 = o1.getStatus()
+    status2 = o2.getStatus()
 
     assert 'version' in status1
     assert 'log_len' in status2
@@ -1408,6 +1467,92 @@ def test_syncobjAdminAddRemove():
     o2._destroy()
 
 
+def test_journalWithAddNodes():
+    dumpFiles = [getNextDumpFile(), getNextDumpFile(), getNextDumpFile()]
+    journalFiles = [getNextJournalFile(), getNextJournalFile(), getNextJournalFile()]
+    removeFiles(dumpFiles)
+    removeFiles(journalFiles)
+    removeFiles([e + '.meta' for e in journalFiles])
+
+    random.seed(42)
+
+    a = [getNextAddr(), getNextAddr(), getNextAddr()]
+
+    o1 = TestObj(a[0], [a[1]], TEST_TYPE.JOURNAL_1, dumpFile=dumpFiles[0], journalFile=journalFiles[0], dynamicMembershipChange=True)
+    o2 = TestObj(a[1], [a[0]], TEST_TYPE.JOURNAL_1, dumpFile=dumpFiles[1], journalFile=journalFiles[1], dynamicMembershipChange=True)
+    objs = [o1, o2]
+    doTicks(objs, 10, stopFunc=lambda: o1._isReady() and o2._isReady())
+
+    assert o1._getLeader().address in a
+    assert o1._getLeader() == o2._getLeader()
+
+    o1.addValue(150)
+    o2.addValue(200)
+
+    doTicks(objs, 10, stopFunc=lambda: o1.getCounter() == 350 and o2.getCounter() == 350)
+
+    assert o1.getCounter() == 350
+    assert o2.getCounter() == 350
+    doTicks(objs, 2)
+
+
+    trueRes = 'SUCCESS ADD ' + a[2]
+    currRes = {}
+    args = {
+        o1: ['-conn', a[0], '-add', a[2]],
+    }
+    doSyncObjAdminTicks([o1, o2], args, 10.0, currRes, stopFunc=lambda: currRes.get(o1) is not None)
+
+    assert currRes[o1] == trueRes
+
+    o3 = TestObj(a[2], [a[1], a[0]], TEST_TYPE.JOURNAL_1, dumpFile=dumpFiles[2], journalFile=journalFiles[2], dynamicMembershipChange=True)
+
+    doTicks([o1, o2, o3], 10.0, stopFunc=lambda: o1._isReady() and o2._isReady() and o3._isReady())
+
+    assert o1._isReady()
+    assert o2._isReady()
+    assert o3._isReady()
+
+    assert o3.getCounter() == 350
+
+    doTicks(objs, 2)
+
+
+    o1._destroy()
+    o2._destroy()
+    o3._destroy()
+
+    removeFiles(dumpFiles)
+
+    o1 = TestObj(a[0], [a[1]], TEST_TYPE.JOURNAL_1, dumpFile=dumpFiles[0], journalFile=journalFiles[0], dynamicMembershipChange=True)
+    o2 = TestObj(a[1], [a[0]], TEST_TYPE.JOURNAL_1, dumpFile=dumpFiles[1], journalFile=journalFiles[1], dynamicMembershipChange=True)
+    o3 = TestObj(a[2], [a[1], a[0]], TEST_TYPE.JOURNAL_1, dumpFile=dumpFiles[2], journalFile=journalFiles[2], dynamicMembershipChange=True)
+
+    objs = [o1, o2, o3]
+    doTicks(objs, 10, stopFunc=lambda: o1._isReady() and o1.getCounter() == 350 and o3._isReady() and o3.getCounter() == 350)
+
+    assert o1._isReady()
+    assert o3._isReady()
+
+    assert o1.getCounter() == 350
+    assert o3.getCounter() == 350
+
+    o2.addValue(200)
+
+    doTicks(objs, 10, stopFunc=lambda: o1.getCounter() == 550 and o3.getCounter() == 550)
+
+    assert o1.getCounter() == 550
+    assert o3.getCounter() == 550
+
+    o1._destroy()
+    o2._destroy()
+    o3._destroy()
+
+    removeFiles(dumpFiles)
+    removeFiles(journalFiles)
+    removeFiles([e + '.meta' for e in journalFiles])
+
+
 def test_syncobjAdminSetVersion():
     random.seed(42)
 
